@@ -30,8 +30,42 @@
 #include <assert.h>
 #include <gavl/gavl.h>
 
+#include "config.h"
+
 /* Video conversion */
 
+#ifdef HAVE_MEMALIGN
+/* some systems have memalign() but no declaration for it */
+void * memalign (size_t align, size_t size);
+#else
+/* assume malloc alignment is sufficient */
+#define memalign(align,size) malloc (size)
+#endif
+
+#define ALIGNMENT_BYTES 16
+#define ALIGN(a) a=((a+ALIGNMENT_BYTES-1)/ALIGNMENT_BYTES)*ALIGNMENT_BYTES
+
+static inline int caml_gavl_video_frame_is_aligned(gavl_video_frame_t *vf, int planes)
+{
+  int i;
+
+  for (i = 0; i < planes; i++)
+  {
+    if (((long)vf->planes[i] % ALIGNMENT_BYTES != 0))
+      return 0;
+    if (((long)vf->strides[i] % ALIGNMENT_BYTES != 0))
+      return 0;
+  }
+
+  return 1;
+}
+
+static inline int caml_gavl_num_planes(gavl_video_format_t *vf)
+{
+  gavl_pixelformat_t pf = vf->pixelformat;
+  return gavl_pixelformat_num_planes(pf);
+}
+
 CAMLprim value caml_gavl_vid_int_of_define(value d)
 {
   CAMLparam1(d);
@@ -330,15 +364,16 @@ static inline int caml_gavl_plane_size(gavl_video_format_t *format, int plane, i
 
 static gavl_video_frame_t *caml_gavl_alloc_frame(gavl_video_frame_t *f, gavl_video_format_t *vf)
 {
-  int p = gavl_pixelformat_num_planes (vf->pixelformat);
+  int p = caml_gavl_num_planes(vf);
   int i,len;
   for (i = 0; i < p; i++)
   {
     len = caml_gavl_bytes_per_line(vf,i);
-    f->planes[i]  = malloc(caml_gavl_plane_size(vf,i,len));
+    ALIGN(len);
+    f->planes[i]  = memalign(ALIGNMENT_BYTES,caml_gavl_plane_size(vf,i,len));
     f->strides[i] = len;
     if (f->planes[i] == NULL)
-      caml_failwith("malloc");
+      caml_raise_out_of_memory();
   }
   f->user_data = NULL;
   f->timestamp = 0;
@@ -355,9 +390,8 @@ static gavl_video_frame_t *gavl_video_frame_of_value(value v, gavl_video_format_
   int j,len,stride;
   struct caml_ba_array *data;
   value tmp,plane;
-  gavl_pixelformat_t pf = vf->pixelformat;
   value planes = Field(v, i++);
-  int np = gavl_pixelformat_num_planes(pf);
+  int np = caml_gavl_num_planes(vf);
   if (np != Wosize_val(planes))
     caml_raise_constant(*caml_named_value("caml_gavl_invalid_frame"));
 
@@ -392,8 +426,7 @@ static value value_of_gavl_video_frame(gavl_video_format_t *vf, gavl_video_frame
   int i = 0;
   int j;
   intnat len;
-  gavl_pixelformat_t pf = vf->pixelformat;
-  int np = gavl_pixelformat_num_planes(pf);
+  int np = caml_gavl_num_planes(vf);
   planes = caml_alloc_tuple(np);
   for (j = 0; j < np; j++)
   {
@@ -450,7 +483,7 @@ CAMLprim value caml_gavl_vid_conv_create(value old, value new)
   int pass;
   vid_conv_t *conv = malloc(sizeof(vid_conv_t));
   if (conv == NULL) 
-    caml_failwith("malloc");
+    caml_raise_out_of_memory();
 
   conv->conv = gavl_video_converter_create();
   if (conv->conv == NULL)
@@ -484,7 +517,9 @@ CAMLprim value caml_gavl_vid_conv_init(value conv, value old, value new)
 
   if (pass == -1)
     caml_raise_constant(*caml_named_value("caml_gavl_invalid_conversion"));
-  
+
+  vid_conv->pass = pass; 
+ 
   CAMLreturn(Val_unit);
 }
 
@@ -502,6 +537,7 @@ CAMLprim value caml_gavl_vid_conv_get_formats(value conv)
 
 CAMLprim value caml_gavl_vid_conv_convert(value conv, value old, value new) 
 {
+  CAMLparam3(conv,old,new);
   vid_conv_t *vid_conv = Vid_conv_val(conv);
 
   /* pass < 0 should not happen since it 
@@ -510,29 +546,52 @@ CAMLprim value caml_gavl_vid_conv_convert(value conv, value old, value new)
 
   gavl_video_converter_t *cnv = vid_conv->conv;
   gavl_video_frame_t inf;
+  gavl_video_frame_t *inft;
+  int in_al = 1;
+
   gavl_video_frame_t outf;
+  gavl_video_frame_t *outft;
+  int out_al = 1;
+
   int j;
+
   gavl_video_frame_of_value(old,&vid_conv->in_vf,&inf);
+  if (caml_gavl_video_frame_is_aligned(&inf,caml_gavl_num_planes(&vid_conv->in_vf)) == 0)
+  {
+    inft = gavl_video_frame_create(&vid_conv->in_vf);
+    gavl_video_frame_copy(&vid_conv->in_vf,inft,&inf);
+    in_al = 0;
+  } else
+    inft = &inf;
+
   gavl_video_frame_of_value(new,&vid_conv->out_vf,&outf);
-
-  /* pass == 0 means no conversion is needed.. */ 
-  if (vid_conv->pass == 0)
-    caml_raise_constant(*caml_named_value("caml_gavl_no_conversion_needed"));
-
-  caml_register_global_root(&old);
-  caml_register_global_root(&new);
-  caml_register_global_root(&conv);
+  if (caml_gavl_video_frame_is_aligned(&outf,caml_gavl_num_planes(&vid_conv->out_vf)) == 0)
+  {
+    outft = gavl_video_frame_create(&vid_conv->out_vf);
+    gavl_video_frame_copy(&vid_conv->out_vf,outft,&outf);
+    out_al = 0;
+  } else
+    outft = &outf;
 
   caml_enter_blocking_section();
-  for (j = 0; j < vid_conv->pass; j++)
-    gavl_video_convert(cnv,&inf,&outf);
+  /* pass == 0 means no conversion is needed.. */
+  if (vid_conv->pass == 0)
+    gavl_video_frame_copy(&vid_conv->in_vf,&outf,inft);
+  else
+  {
+    for (j = 0; j < vid_conv->pass; j++)
+      gavl_video_convert(cnv,inft,outft);
+    if (out_al == 0)
+      gavl_video_frame_copy(&vid_conv->out_vf,&outf,outft);
+  }
   caml_leave_blocking_section();  
 
-  caml_remove_global_root(&old);
-  caml_remove_global_root(&new);
-  caml_remove_global_root(&conv);
+  if (in_al == 0)
+    gavl_video_frame_destroy(inft);
+  if (out_al == 0)
+    gavl_video_frame_destroy(outft);
 
-  return Val_unit;
+  CAMLreturn(Val_unit);
 }
 
 CAMLprim value caml_gavl_vid_conv_new_frame(value format)
@@ -607,7 +666,13 @@ CAMLprim value caml_gavl_vid_conv_reinit(value conv)
 {
   CAMLparam1(conv);
   vid_conv_t *vid_conv = Vid_conv_val(conv);
-  gavl_video_converter_reinit(vid_conv->conv);
+  int pass = gavl_video_converter_reinit(vid_conv->conv);
+
+  if (pass == -1)
+    caml_raise_constant(*caml_named_value("caml_gavl_invalid_conversion"));
+
+  vid_conv->pass = pass ;
+
   CAMLreturn(Val_unit);
 }
 
